import enum


class ImagePullPolicy(enum.Enum):
    IfNotPresent = "IfNotPresent"
    Always = "Always"
    Never = "Never"


METAAI_USER_INFO_PLACEHOLDER = "{{workflow.annotations.metaai.zjuici.com/user-id}}"


METAAI_MLMODELS_ENDPOINT_PLACEHOLDER = (
    "{{workflow.annotations.metaai.zjuici.com/mlmodels-endpoint}}"
)

METAAI_MLDATASETS_ENDPOINT_PLACEHOLDER = (
    "{{workflow.annotations.metaai.zjuici.com/mldatasets-endpoint}}"
)



IMAGE_REPO_HOST = "10.0.0.68:8089"

DATASETS_IMAGE = f"{IMAGE_REPO_HOST}/metaai/metaai-client:0.2.0"

MODELS_IMAGE = f"{IMAGE_REPO_HOST}/metaai/metaai-client:0.2.0"

# 配置harbor的python，使用清华镜像的源
DEFAULT_BASE_PYTHON_IMAGE = f"{IMAGE_REPO_HOST}/library/python:3.7"

KFP_DISABLE_CACHE_ANNOTATIONS = {"pipelines.kubeflow.org/max_cache_staleness": "P0D"}



from kubernetes import client


def use_shared_memory_to_volume(op):
    name = "dhsm"
    op.add_volume(
        client.V1Volume(
            name=name, empty_dir=client.V1EmptyDirVolumeSource(medium="Memory")
        )
    ).add_volume_mount(client.V1VolumeMount(name=name, mount_path="/dev/shm"))
    return op



import os


DEFAULT_MODEL_FOLDER_PATH = os.getenv("MODEL_FOLDER_PATH", "/mnt/models")

DEFAULT_MODEL_NAME = os.getenv("MODEL_NAME", "model")

DEFAULT_MODEL_PATH = os.getenv(
    "MODEL_PATH", str(os.path.join(DEFAULT_MODEL_FOLDER_PATH, DEFAULT_MODEL_NAME))
)



import os
from fastapi import FastAPI, APIRouter, Header, Request
from fastapi.responses import JSONResponse
from .model import Model
from .exceptions import ServingHttpError
import uvicorn


def authn(user_id):

    if not user_id or user_id != os.getenv("KUBEFLOW_USER_ID"):
        return False

    return True


class ModelServer:
    @classmethod
    def create_application(cls) -> FastAPI:
        # return APIRouter(redirect_slashes=True)
        return FastAPI(redoc_url=None, openapi_url=None)

    @classmethod
    def _add_error_handler(cls, app: FastAPI):
        @app.exception_handler(ServingHttpError)
        async def serving_error_handler(request, exc: ServingHttpError):
            return JSONResponse(status_code=exc.status_code, content=exc.detail)

    @classmethod
    def _include_api_route(cls, app: FastAPI, model: Model):
        api = APIRouter(redirect_slashes=True)

        @api.post("/predict/")
        async def predict(body: dict, kubeflow_userid: str = Header(default=None)):
            if not authn(kubeflow_userid):

                return JSONResponse(
                    status_code=401,
                    content={
                        "code": 401,
                        "message": f"kubeflow-userid {kubeflow_userid} not found in our records",
                    },
                )

            return await model(body=body)

        app.include_router(api)

    @classmethod
    def start(cls, model: Model):

        if not isinstance(model, Model):
            raise ValueError(
                "Please inherit 'cli.py.serving.models.ModelService' for custom model service class"
            )
        try:
            model.load()
        except Exception as load_e:
            raise RuntimeError(f"loading model file were error ! error msg {load_e}")

        if not model.ready:
            raise RuntimeError(
                "ModelService load model not ready! check you class,please!}"
            )
        app = cls.create_application()

        @app.get("/")
        async def health():
            return {"status": "Alive"}

        cls._include_api_route(app, model)
        cls._add_error_handler(app)
        # 使用命令行运行服务，可能会出现的问题
        # 加载模型的时候，一些类，需要在uvicorn文件上显式的声明 这显然不友好

        # 使用当前方法 目前看来和命令行启动并没有太多的不一样
        uvicorn.run(app, host="0.0.0.0", port=8089, proxy_headers=True)



from typing import *
from kfp.components import _structures as structures
from kfp.components._structures import CommandlineArgumentType


class HackContainerSpec(structures.ModelBase):
    _serialized_names = {
        "file_outputs": "fileOutputs",
        # "image_pull_policy": "imagePullPolicy",
    }

    def __init__(
        self,
        image: str,
        command: Optional[List[CommandlineArgumentType]] = None,
        args: Optional[List[CommandlineArgumentType]] = None,
        env: Optional[Mapping[str, str]] = None,
        file_outputs: Optional[Mapping[str, str]] = None,
        # image_pull_policy: str = None,
    ):

        super().__init__(locals())


class HackContainerImplementation(structures.ModelBase):
    def __init__(
        self,
        container: HackContainerSpec,
    ):
        super().__init__(locals())



from typing import *
from kfp.dsl import pipeline as dsl_pipeline
import kfp.compiler
from kfp.dsl import PipelineConf

from metaai.pipeline.components.constants import ImagePullPolicy

__all__ = ["pipeline"]


class Pipeline:
    def __init__(self, func: Callable):
        self.func = func
        self._pipeline_conf = PipelineConf()
        self._pipeline_func = None

    def _set_image_pull_policy(self, value: str):
        if not value:
            return
        if value not in [
            ImagePullPolicy.Always.value,
            ImagePullPolicy.IfNotPresent.value,
            ImagePullPolicy.Never.value,
        ]:
            raise ValueError(
                "Invalid image_pull_policy. Must be one of `Always`, `Never`, `IfNotPresent`,"
                "see enum of metaai.pipeline.constants.ImagePullPolicy"
            )
        self._pipeline_conf.set_image_pull_policy(value)

    def __call__(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        image_pull_policy: Optional[str] = None,
    ):
        self._pipeline_func = dsl_pipeline(name=name, description=description)(
            self.func
        )

        self._set_image_pull_policy(image_pull_policy)
        return self

    def to_save(self, file_name: str, file_type: str = "yaml"):
        kfp.compiler.Compiler().compile(
            self.func, f"{file_name}.{file_type}", pipeline_conf=self._pipeline_conf
        )

    def to_yaml(self, file_name):
        self.to_save(file_name=file_name, file_type="yaml")


def pipeline(
    name: Optional[str] = None,
    description: Optional[str] = None,
    image_pull_policy: Optional[str] = None,
):
    def wrapper(func):
        return Pipeline(func)(
            name=name, description=description, image_pull_policy=image_pull_policy
        )

    return wrapper



from typing import Dict, Optional
from kfp.components import OutputPath

from ... import settings
from ..component_decorator import component
from .. import set_required_env_in_op_from_workflow


def tabular_datasets(
    datasets_repo: Dict,
    enable_cache: Optional[bool] = True,
):
    @component(
        base_image=settings.DATASETS_IMAGE,
        enable_cache=enable_cache,
    )
    def tabular_datasets_op(datasets_origin: Dict, datasets_path: OutputPath(str)):
        from metaaiclient import DatasetClient

        data_client = DatasetClient(datasets_origin, datasets_path)

        data_client(datasets_type="tabular")

    return set_required_env_in_op_from_workflow(tabular_datasets_op(datasets_repo))


def image_datasets(
    datasets_repo: Dict,
    enable_cache: Optional[bool] = True,
):
    # 新增env
    @component(
        base_image=settings.DATASETS_IMAGE,
        enable_cache=enable_cache,
    )
    def image_datasets_op(datasets_origin: Dict, datasets_path: OutputPath(str)):
        from metaaiclient import DatasetClient

        data_client = DatasetClient(datasets_origin, datasets_path)

        data_client(datasets_type="image")

    return set_required_env_in_op_from_workflow(image_datasets_op(datasets_repo))



from fastapi.responses import JSONResponse
from .models.commons.exceptions import ServingHttpError


def __mk_response_body(code, message, data):
    return {"code": code, "message": message, "data": data}


def _mk_json_response(code, message, data, status_code):
    return JSONResponse(
        __mk_response_body(
            code,
            message,
            data,
        ),
        status_code=status_code,
    )


def success_response(body: dict, status_code: int = 200):
    if not isinstance(body, dict):
        raise ValueError("body must be a dict!")
    return _mk_json_response(0, "Success", body, status_code)


def failed_response(body: dict = None, message: str = None, status_code: int = 400):
    if body is None:
        body = {}
    if not isinstance(body, dict):
        raise ValueError("body must be a dict!")
    return _mk_json_response(-1, message if message else "Failed", body, status_code)


def failed_response_with_break(
    body: dict = None, message: str = None, status_code: int = 400
):
    if body is None:
        body = {}
    if not isinstance(body, dict):
        raise ValueError("body must be a dict!")

    raise ServingHttpError(
        status_code=status_code,
        detail=__mk_response_body(-1, message if message else "Failed", body),
    )


__all__ = ["success_response", "failed_response"]



from ... import settings
from ..component_decorator import component
from kfp.components import OutputPath


# metaai oss location
def download_model_by_url(base_model_url: str, enable_cache: bool = False):
    @component(
        base_image=settings.MODELS_IMAGE,
        enable_cache=enable_cache,
    )
    def download_base_model_op(model_url: str, base_model_path: OutputPath(str)):
        from metaaiclient import ModelClient

        ModelClient(
            model_url,
            {},
            None,
            output=base_model_path,
        ).download_base_model()

    return download_base_model_op(base_model_url)



from typing import *
import functools
from metaai.pipeline.components.hack import create_component_from_func
from metaai.pipeline.components import settings


# see kfp.v2.components.component_decorator component
def component(
    func: Optional[Callable] = None,
    *,
    enable_cache: Optional[bool] = True,
    output_component_file: Optional[str] = None,
    base_image: Optional[str] = settings.DEFAULT_BASE_PYTHON_IMAGE,
    packages_to_install: List[str] = None,
    # image_pull_policy: Optional[str] = None,
    annotations: Optional[Mapping[str, str]] = None,
):
    if func is None:
        return functools.partial(
            component,
            enable_cache=enable_cache,
            base_image=base_image,
            packages_to_install=packages_to_install,
            output_component_file=output_component_file,
            # image_pull_policy=image_pull_policy,
            annotations=annotations,
        )

    return create_component_from_func(
        func,
        enable_cache=enable_cache,
        output_component_file=output_component_file,
        base_image=base_image,
        packages_to_install=packages_to_install,
        # image_pull_policy=image_pull_policy,
        annotations=annotations,
    )



from typing import Any, Dict
import inspect
from ... import settings


class Model:
    def __init__(self, name: str):
        self.name = name

        self.ready = False
        self.model_path = settings.DEFAULT_MODEL_PATH

    async def _preprocess(self, request: Any) -> Any:
        return request

    def _postprocess(self, response: Any) -> Dict:
        return response

    async def predict(self, request: Any):
        raise NotImplementedError

    def load(self):
        self.ready = True
        
        return self.ready

    def _validate(self, request: Any) -> Any:
        return request

    async def __call__(self, body: dict, *args, **kwargs):
        request = (
            await self._preprocess(body)
            if inspect.iscoroutinefunction(self._preprocess)
            else self._preprocess(body)
        )
        request = self._validate(request)

        response = (
            (await self.predict(request))
            if inspect.iscoroutinefunction(self.predict)
            else self.predict(request)
        )

        return self._postprocess(response)



from kfp.dsl import ContainerOp

from kubernetes.client import V1EnvVar
from .. import constants


def set_required_env_in_op_from_workflow(op: ContainerOp):
    # op.add_env_variable() 这个也可以，使用下面这个好定位，以及配合 type hint
    op.container.add_env_variable(
        V1EnvVar(name="METAAI_USER_ID", value=constants.METAAI_USER_INFO_PLACEHOLDER)
    )
    op.container.add_env_variable(
        V1EnvVar(
            name="MLMODELS_ENDPOINT",
            value=constants.METAAI_MLMODELS_ENDPOINT_PLACEHOLDER,
        )
    )
    op.container.add_env_variable(
        V1EnvVar(
            name="DATASETS_ENDPOINT",
            value=constants.METAAI_MLDATASETS_ENDPOINT_PLACEHOLDER,
        )
    )

    return op



from typing import *
from kfp.components._python_op import _func_to_component_dict
from kfp.components._python_op import _create_task_factory_from_component_spec
from kfp.components.structures import ComponentSpec
from kfp.components import _structures as structures

from ..settings import KFP_DISABLE_CACHE_ANNOTATIONS


def create_component_from_func(
    func: Callable,
    enable_cache: Optional[bool] = True,
    output_component_file: Optional[str] = None,
    base_image: Optional[str] = None,
    packages_to_install: List[str] = None,
    # image_pull_policy: Optional[str] = None,
    annotations: Optional[Dict[str, str]] = None,
):
    component_spec_dict = _func_to_component_dict(
        func, base_image=base_image, packages_to_install=packages_to_install
    )
    # NOTE: 这个东西写到component的yaml里面了 然后发现load不进来，不支持这个参数！。简直离谱！
    # container_implementation = component_spec_dict.pop("implementation")
    #
    # container_dict = container_implementation.pop("container")
    # NOTE: 主要添加了当前这个镜像的拉取策略

    # if image_pull_policy is not None:
    #     try:
    #         ImagePullPolicy(image_pull_policy)
    #     except Exception:
    #         raise RuntimeError(
    #             f"image_pull_policy now supported value from metaai.pipeline.constants.ImagePullPolicy,"
    #             f"given '{image_pull_policy}' not in enumeration"
    #         )
    #     # NOTE: 这个dict的key是和在yaml中的key相同的
    #     container_dict["imagePullPolicy"] = image_pull_policy

    component_spec = ComponentSpec.from_dict(component_spec_dict)

    # component_spec.implementation = HackContainerImplementation(
    #     container=HackContainerSpec.from_dict(container_dict)
    # )

    if not annotations:
        annotations = {}

    if not enable_cache:

        annotations.update(KFP_DISABLE_CACHE_ANNOTATIONS)
        component_spec.metadata = structures.MetadataSpec(
            annotations=annotations,
        )

    if output_component_file:
        component_spec.save(output_component_file)

    return _create_task_factory_from_component_spec(component_spec)



# !/usr/bin/python3
import argparse
from argparse import Namespace
import os
import subprocess
from urllib.parse import urlparse
import sys
from string import Template
import warnings
import tempfile

from . import __version__


def format_str_with_character(s: str):
    max_str_len = 120

    prefix = f'{"+" * 20} {s} '

    suffix_character_len = max_str_len - len(prefix) - 1

    return f'{prefix} {"+" * suffix_character_len}'


def src_dir_path(path):
    if os.path.isdir(path):
        abs_path = os.path.abspath(path)
        files = os.listdir(path)
        if "serving.py" not in files:
            raise argparse.ArgumentTypeError(
                f"serving.py file not found in dir path {abs_path}"
            )

        if "requirements.txt" not in files:
            raise argparse.ArgumentTypeError(
                f"requirements.txt file not found in dir path {abs_path}"
            )

        return abs_path
    else:
        raise argparse.ArgumentTypeError(f"readable_dir:{path} is not a valid path")


dockerfile_string_template = Template(
    """
# 版本跟随 metaai
FROM --platform=linux/arm64 registry.cn-hangzhou.aliyuncs.com/metaai/python-metaai:${metaai_version} as dependencies

ARG PYPI_INDEX="https://pypi.tuna.tsinghua.edu.cn/simple"
ARG TRUSTED_HOST="pypi.tuna.tsinghua.edu.cn"

WORKDIR /build

COPY ./requirements.txt ./

RUN python3 -m pip config set global.index-url $PYPI_INDEX \
  && python3 -m pip config set global.trusted-host $TRUSTED_HOST \
  && python3 -m pip install --upgrade pip --no-cache-dir \
  && python3 -m pip install -r ./requirements.txt --no-cache-dir \
  && rm /build -rf


FROM dependencies as app

WORKDIR /app

COPY . /app


EXPOSE 8089

CMD ["python3", "serving.py"]
"""
)


def main():
    print(
        format_str_with_character(f"current metaai cli script version is {__version__}")
    )

    dockerfile_string = dockerfile_string_template.safe_substitute(
        {"metaai_version": __version__}
    )

    parser = argparse.ArgumentParser(
        prog="METAAI",
        usage="cli script for metaai",
        description="Provides some command-line functions of the service metaai",
    )

    sub_parser = parser.add_subparsers(
        title="metaai cli subcommands",
    )

    serving_build = sub_parser.add_parser(
        "serving-build",
        help="build serving predict image and batch predict image",
    )

    serving_build.add_argument(
        "-t",
        dest="tag",
        type=str,
        required=True,
        help="build docker image's tag",
    )

    serving_build.add_argument(
        "-s",
        dest="src_path",
        required=True,
        type=src_dir_path,
        help="source code path",
    )

    serving_build.add_argument(
        "--pypi_index",
        dest="pypi_index",
        help="pypi index for pip install，default use 'pypi.tuna.tsinghua.edu.cn' ",
    )

    known_args, _ = parser.parse_known_args()
    if known_args == Namespace():
        parser.parse_args(["--help"])

    # check docker version
    print(format_str_with_character("checking docker version"))

    docker_version = subprocess.Popen(
        "docker version -f {{.Server.Version}}",
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    min_docker_version_required = (20, 10, 14)

    try:
        local_docker_version = (
            docker_version.stdout.readline().decode("utf-8").splitlines()[0]
        )

        print(
            format_str_with_character(
                f"local current docker server version: {local_docker_version}"
            )
        )

    except Exception as e:
        raise SystemExit(f"parser docker version raise error,msg {e}")
    else:
        local_docker_version = tuple(int(i) for i in local_docker_version.split("."))
        if min_docker_version_required > local_docker_version:
            str_version = ".".join(str(i) for i in min_docker_version_required)
            warnings.warn(f"recommended that docker version be gte {str_version}")

    pypi_index = known_args.pypi_index

    build_args_command = ""
    if pypi_index:
        build_args_command = (
            f"--build-arg PYPI_INDEX={pypi_index} "
            f"--build-arg TRUSTED_HOST={urlparse(pypi_index).netloc}"
        )

    tmp_fp = tempfile.NamedTemporaryFile()
    tmp_fp.write(dockerfile_string.encode())
    tmp_fp.seek(0)

    docker_build_cmd_string = (
        f"docker build -t {known_args.tag} {build_args_command}"
        f"-f {tmp_fp.name} {known_args.src_path}"
    )
    print(docker_build_cmd_string)

    docker_builder = subprocess.Popen(
        docker_build_cmd_string,
        shell=True,
        stdout=sys.stdout,
        stderr=sys.stderr,
    )
    docker_builder.communicate()


if __name__ == "__main__":
    main()



from typing import Dict, Optional

from kfp.dsl import RUN_ID_PLACEHOLDER
from kfp.components import InputPath

from ... import settings
from ..component_decorator import component
from .. import set_required_env_in_op_from_workflow


def upload_model(
    model: InputPath(str),
    to_upload_model: Dict,
    metadata_schema: Dict,
    enable_cache: Optional[bool] = True,
):
    @component(
        base_image=settings.MODELS_IMAGE,
        enable_cache=enable_cache,
    )
    def upload_model_op(
        model_path: InputPath(str),
        model_info: Dict,
        meta_schema: Dict,
        run_id: str = None,
    ):

        from metaaiclient import ModelClient

        ModelClient(
            model_path,
            model_info,
            run_id,
            metadata_schema=meta_schema,
        )()

    return set_required_env_in_op_from_workflow(
        upload_model_op(
            model,
            to_upload_model,
            metadata_schema,
            RUN_ID_PLACEHOLDER,
        )
    )
